﻿/*********************************************************************
 *  SPDX-License-Identifier: MIT									 *
 *  Copyright (C) 2019-2020 by Helge Kortendieck  alan@kasmithee.de  *
 *                                                                   *
 *  This file is part of qublc.                                      *
 *********************************************************************/

#include "qublc.h"

Qublc::Qublc()
{
	bAhw = false;							// not showing about/error/help dialog
	bListChanged = false;					// list was not changed yet, no need to ask to save on exit
	iCurCol = 0;
	iCurRowIx = 0;
	liDomains.clear();
	sFileDomain = "";						// current domain list file name (path+name)
	sFileUnbound = "";						// current unbound blacklist.conf file name (path+name)

	uiublcw.setupUi(&wUblcw);
	setCentralWidget(&wUblcw);
	uiublcw.l_DomainFile->setText("");

	create_Actions();
	create_Menus();
	create_ToolBars();
	create_StatusBar();
	create_TableMenu();

	settings_Read();
	table_HeaderSet();

	uiublcw.tW_List->setContextMenuPolicy(Qt::CustomContextMenu);																// needed for right-click context menu
	connect(uiublcw.tW_List, SIGNAL(customContextMenuRequested(const QPoint&)), this, SLOT(table_MenuCalled(const QPoint&)));	// right-click menu on table widget - add new or delete selected item
	connect(uiublcw.tW_List, SIGNAL(cellChanged(int, int)), this, SLOT(table_CellChanged(int, int)));							// row, col: item in table widget changed, update also the corresponding domain struct list field
	connect(uiublcw.tW_List->horizontalHeader(), SIGNAL(sectionResized(int, int, int)), this, SLOT(timer5SaveHeadWidth()));		// 5sec delay: check col headers w resized->save
	connect(uiublcw.tW_List->horizontalHeader(), SIGNAL(sectionClicked(int)), this, SLOT(table_HeaderClicked(int)));			// a table header was clicked - anything to do?   int = column clicked

	timer5ListHead = new QTimer(this);			// 5 sec timer to limit checks if the table column headers where resized
	timer5ListHead->setSingleShot(true);
	connect(timer5ListHead, SIGNAL(timeout()), this, SLOT(settings_Write()));	// save new column width of table widget

	addDomainDlg_Create();

	QScreen *screen = QGuiApplication::primaryScreen();
	qrFrame = screen->geometry();
	iScrX = qrFrame.width();					// screen size horizontal (pixels)
	iScrY = qrFrame.height();					// screen size vertical (pixels)
}

Qublc::~Qublc()
{
}

// ********************************************************
// **  create actions, menu, tollbar, statusbar			 **
// ********************************************************
void Qublc::create_Actions()
{
// *****  program actions
	aboutAct = new QAction(QIcon( "/usr/share/qublc/icons/scalable/about.svg"), tr("a&Bout qublc"), this);
	aboutAct->setShortcut(tr("Ctrl+B"));
	aboutAct->setStatusTip(tr("Show the application's About dialog"));
	aboutAct->setToolTip(tr("Show About dialog"));
	connect(aboutAct, SIGNAL(triggered()), this, SLOT(about()));

	helpAct = new QAction(QIcon( "/usr/share/qublc/icons/scalable/help.svg"), tr("&Help"), this);
	helpAct->setShortcut(tr("Ctrl+H"));
	helpAct->setStatusTip(tr("Show the application's online help"));
	helpAct->setToolTip(tr("Show online help"));
	connect(helpAct, SIGNAL(triggered()), this, SLOT(help()));

	exitAct = new QAction(QIcon( "/usr/share/qublc/icons/scalable/exit.svg" ), tr("&Quit the program"), this);
	exitAct->setShortcut(tr("Ctrl+Q"));
	exitAct->setStatusTip(tr("Quit the application"));
	exitAct->setToolTip(tr("Quit qublc"));
	connect(exitAct, SIGNAL(triggered()), this, SLOT(quit_Prep()));

// *****  domain list file actions
	domainListNewAct = new QAction(QIcon( "/usr/share/qublc/icons/scalable/new.svg" ), tr("create &New domain list"), this);
	domainListNewAct->setShortcut(tr("Ctrl+N"));
	domainListNewAct->setStatusTip(tr("Create a new, empty domain list"));
	domainListNewAct->setToolTip(tr("Create new, empty domain list"));
	connect(domainListNewAct, SIGNAL(triggered()), this, SLOT(domainListFile_New()));

	domainListOpenAct = new QAction(QIcon( "/usr/share/qublc/icons/scalable/open.svg" ), tr("&Open existing domain list..."), this);
	domainListOpenAct->setShortcut(tr("Ctrl+O"));
	domainListOpenAct->setStatusTip(tr("Open an existing domain list"));
	domainListOpenAct->setToolTip(tr("Open existing domain list"));
	connect(domainListOpenAct, SIGNAL(triggered()), this, SLOT(domainListFile_OpenDlg()));

	domainListSaveAct = new QAction(QIcon( "/usr/share/qublc/icons/scalable/save.svg" ), tr("&Save domain list"), this);
	domainListSaveAct->setShortcut(tr("Ctrl+S"));
	domainListSaveAct->setStatusTip(tr("Save domain list"));
	domainListSaveAct->setToolTip(tr("Save domain list"));
	connect(domainListSaveAct, SIGNAL(triggered()), this, SLOT(domainListFile_Save()));

	domainListSaveAsAct = new QAction(QIcon( "/usr/share/qublc/icons/scalable/save-as.svg" ), tr("save domain list &As..."), this);
	domainListSaveAsAct->setShortcut(tr("Ctrl+A"));
	domainListSaveAsAct->setStatusTip(tr("Save this domain list under a new name"));
	domainListSaveAsAct->setToolTip(tr("Save this domain list under a new name"));
	connect(domainListSaveAsAct, SIGNAL(triggered()), this, SLOT(domainListFile_SaveAs()));

// *****  domain list widget actions
	domainListDomainAddAct = new QAction(QIcon( "/usr/share/qublc/icons/scalable/plus.svg" ), tr("add new domain entry (&P)"), this);
	domainListDomainAddAct->setShortcut(tr("Ctrl+P"));
	domainListDomainAddAct->setStatusTip(tr("Add a new domain entry"));
	domainListDomainAddAct->setToolTip(tr("Add a new domain entry"));
	connect(domainListDomainAddAct, SIGNAL(triggered()), this, SLOT(addDomainDlg_Show()));

	domainListDomainDelAct = new QAction(QIcon( "/usr/share/qublc/icons/scalable/minus.svg" ), tr("&Delete current domain entry"), this);
	domainListDomainDelAct->setShortcut(tr("Ctrl+D"));
	domainListDomainDelAct->setStatusTip(tr("Delete this domain entry"));
	domainListDomainDelAct->setToolTip(tr("Delete this domain entry"));
	connect(domainListDomainDelAct, SIGNAL(triggered()), this, SLOT(domainListEntry_Del()));

// *****  unbound blocking conf actions
	unboundFileCreateAct = new QAction(QIcon( "/usr/share/qublc/icons/scalable/export.svg" ), tr("Create &Unbound blocking configuration file"), this);
	unboundFileCreateAct->setShortcut(tr("Ctrl+E"));
	unboundFileCreateAct->setStatusTip(tr("Export this domain list as a blocking configuration file for unbound"));
	unboundFileCreateAct->setToolTip(tr("Export this domain list as a blocking configuration file for unbound"));
	connect(unboundFileCreateAct, SIGNAL(triggered()), this, SLOT(unboundFile_Create()));
}

void Qublc::create_Menus()
{
	fileMenu = menuBar()->addMenu(tr("&File"));
	fileMenu->addAction(domainListNewAct);
	fileMenu->addSeparator();
	fileMenu->addAction(domainListOpenAct);
	fileMenu->addSeparator();
	fileMenu->addAction(domainListSaveAct);
	fileMenu->addAction(domainListSaveAsAct);
	fileMenu->addSeparator();
	fileMenu->addAction(exitAct);

	menuBar()->addSeparator();

	exportMenu = menuBar()->addMenu(tr("&Export"));
	exportMenu->addAction(unboundFileCreateAct);

	menuBar()->addSeparator();

	helpMenu = menuBar()->addMenu(tr("&Help"));
	helpMenu->addAction(helpAct);
	helpMenu->addAction(aboutAct);
}

void Qublc::create_ToolBars()
{
	fileToolBar = addToolBar(tr("Show toolbar"));
	fileToolBar->addAction(domainListNewAct);
	fileToolBar->addAction(domainListOpenAct);
	fileToolBar->addAction(domainListSaveAct);
	fileToolBar->addAction(domainListSaveAsAct);
	fileToolBar->addSeparator();
	fileToolBar->addAction(unboundFileCreateAct);
	fileToolBar->addSeparator();

	fileToolBar->addAction(exitAct);
/*
	fileToolBar->addAction(mCompareAct);
	fileToolBar->addAction(mDelRadioAct);
	fileToolBar->addSeparator();
	fileToolBar->addAction(cOpenAct);
	fileToolBar->addAction(cDelRadioAct);
	fileToolBar->addAction(cDelTvAct);
*/
}

void Qublc::create_StatusBar()
{
	statusBar()->showMessage(tr("Ready"));
}

void Qublc::create_TableMenu()				// domain list widget right-click menu - add/delete entry
{
	menuDomainListRightClick = new QMenu;
	menuDomainListRightClick->clear();
	menuDomainListRightClick->addAction(domainListDomainAddAct);
	menuDomainListRightClick->addAction(domainListDomainDelAct);
}

// ********************************************************
// **  read/write settings                               **
// ********************************************************
void Qublc::settings_Read()
{
	QSettings settings("qublc", "qublc");
	settings.beginGroup( "main" );
		qPos = settings.value("pos", QPoint(200, 200)).toPoint();
		size = settings.value("size", QSize(400, 400)).toSize();
		sFileDomainLast = settings.value("filedomain", "/tmp/cur.qublc").toString();
		sFileUnboundLast = settings.value("fileunbound", "/tmp/block.conf").toString();
		iColWidth0 = settings.value("listwidthcol0", 50).toInt(); // list widget: width column 0
		iColWidth1 = settings.value("listwidthcol1", 50).toInt(); // list widget: width column 1
		iColWidth2 = settings.value("listwidthcol2", 50).toInt(); // list widget: width column 2
		iColWidth3 = settings.value("listwidthcol3", 50).toInt(); // list widget: width column 3
		iColWidth4 = settings.value("listwidthcol4", 50).toInt(); // list widget: width column 4
		iColWidth5 = settings.value("listwidthcol5", 50).toInt(); // list widget: width column 5
		iColWidth6 = settings.value("listwidthcol6", 50).toInt(); // list widget: width column 6
		iSortDl = settings.value("sortstartcol", 1).toInt(); // sort settings: 1=2nd+3rd, 2=top+2nd+3rd level domain
	settings.endGroup();

	resize(size);
	move(qPos);
}

void Qublc::settings_Write()
{
	QSettings settings("qublc", "qublc");
	settings.beginGroup("main");
		settings.setValue("pos", QMainWindow::pos());
		settings.setValue("size", QMainWindow::size());
		settings.setValue("filedomain", sFileDomain);
		settings.setValue("fileunbound", sFileUnbound);
		settings.setValue("listwidthcol0", uiublcw.tW_List->columnWidth(0));	// master list widget: width column 0
		settings.setValue("listwidthcol1", uiublcw.tW_List->columnWidth(1));	// master list widget: width column 1
		settings.setValue("listwidthcol2", uiublcw.tW_List->columnWidth(2));	// master list widget: width column 2
		settings.setValue("listwidthcol3", uiublcw.tW_List->columnWidth(3));	// master list widget: width column 3
		settings.setValue("listwidthcol4", uiublcw.tW_List->columnWidth(4));	// master list widget: width column 4
		settings.setValue("listwidthcol5", uiublcw.tW_List->columnWidth(5));	// master list widget: width column 5
		settings.setValue("listwidthcol6", uiublcw.tW_List->columnWidth(6));	// master list widget: width column 6
		settings.setValue("sortstartcol", iSortDl);	// sort settings: 1=2nd+3rd, 2=top+2nd+3rd level domain
	settings.endGroup();

	settings.sync();
}

// ********************************************************
// **  closing/quitting                                  **
// ********************************************************
void Qublc::closeEvent(QCloseEvent *event)
{
	event->accept();
	quit_Prep();
}

void Qublc::quit_Prep()						// quit clicked clean up
{
	if (bListChanged == true) {
		iD = mbShow(5, tr("Question - save changes ?"),
		                   tr("The domain list has been modified.\n \nSave changes before quiting?\n"),
		                   tr("Yes - save domain list"), tr("No - discard changes"), tr("Cancel - do not quit"));
		if (iD == 1) {
			return;
		} else if (iD == 0) {
			if (sFileDomain != "") {
				domainListFile_Save();
			} else {
				domainListFile_SaveAs();
			}
		}
	}
	settings_Write();
	quit_Do();									// only for now...
}

void Qublc::quit_Do()						// finally quit the program
{
	exit(0);
}

// ********************************************************
// **  timers											 **
// ********************************************************
void Qublc::timer5SaveHeadWidth()
{
	if (timer5ListHead->isActive() == false) {
		timer5ListHead->start(5000);
	}
}

// ********************************************************
// **  widget populate / clear		                     **
// ********************************************************
void Qublc::table_CellChangedConnect()		// connect signal: widget cell changed
{
	connect(uiublcw.tW_List, SIGNAL(cellChanged(int, int)), this, SLOT(table_CellChanged(int, int)));
}

void Qublc::table_CellChangedDisConnect()	// disconnect signal: widget cell changed
{
	disconnect(uiublcw.tW_List, SIGNAL(cellChanged(int, int)), this, SLOT(table_CellChanged(int, int)));
}

void Qublc::table_CellChanged(int iD, int iD2)	// row, col: data of a cell has changed, check for '|', double domain entries and update underlying domain list
{
	iCurCol = iD2;
	iCurRow = iD;
	sChk = uiublcw.tW_List->item(iD, iD2)->text();
	slDomain.clear();
	slDomain.append(uiublcw.tW_List->item(iCurRow, 0)->text());
	slDomain.append(uiublcw.tW_List->item(iCurRow, 1)->text());
	slDomain.append(uiublcw.tW_List->item(iCurRow, 2)->text());

	if (domainListEntry_ChkSeperator(sChk) == true) {	// check for '|' (data set field separator), replace with '/'
		sChk.replace("|", "/");
		QTableWidgetItem *newItem = new QTableWidgetItem(sChk);
		if (iCurCol != 4 and iCurCol != 6) {
			newItem->setTextAlignment(Qt::AlignCenter);
		}
		uiublcw.tW_List->setItem(iCurRow, iCurCol, newItem);		// row, col
		mbTxt_DomainListEntry_SepRep();
	}

	if (iCurCol < 3) {									// a domain field has changed - check for double counts
		iD3 = iCurRow;
		iD4 = iCurCol;
		slDomain.replace(iCurCol, sChk);					// update domain from changed cell entry
		if (domainListEntry_FindDomain(slDomain) == true) {	// if true, domain already exists (0=3rd,2=top level)
			mbTxt_DomainListEntry_ChgReverted();
			table_Upd();									// re-populate the table with old values
			uiublcw.tW_List->setCurrentCell(iD3, iD4);
			return;
		}
	}

	if (iCurCol == 3) {									// include field has changed - check for invalid entry
		iD3 = iCurRow;
		iD4 = iCurCol;
		if (sChk != "" or sChk != "x" == true) {			// if true, entry invalid - domain already exists (0=3rd,2=top level)
			table_Upd();									// re-populate the table with old values
			mbTxt_DomainListEntry_InvalidInclude();
			uiublcw.tW_List->setCurrentCell(iD3, iD4);
			return;
		}
	}

	strDomainD = liDomains.at(iCurRow);						// replace value in underlying list
	switch (iCurCol) {
	    case 0:
		    strDomainD.dl3 = sChk;
		    break;
	    case 1:
		    strDomainD.dl2 = sChk;
		    break;
	    case 2:
		    strDomainD.dltop = sChk;
		    break;
	    case 3:
		    strDomainD.include = sChk;
		    break;
	    case 4:
		    strDomainD.note = sChk;
		    break;
	    case 5:
		    strDomainD.lastchkd = sChk;
		    break;
	    case 6:
		    strDomainD.recfrom = sChk;
		    break;
	}
	liDomains.replace(iCurRow, strDomainD);
	if (iCurCol < 3) {										// sort and update widget if domain level field changed
		domainListEntry_Sort();
		table_Upd();
		domainListEntry_FindDomain(slDomain);
		uiublcw.tW_List->setCurrentCell(iDomainListPos, iCurCol);
	}
	bListChanged = true;
}

void Qublc::table_Clear()					// clear widget
{
	uiublcw.tW_List->clearContents();
	uiublcw.tW_List->setRowCount(0);
	iCurRowIx = -1;
	iCurCol = -1;
}

void Qublc::table_MenuCalled(const QPoint& qPD)	// right click in table widget - call context menu to add/delete entry
{
	qPos = qPD;
	menuDomainListRightClick->exec(QCursor::pos());
}

void Qublc::table_Pop()						// populate widget
{
	table_CellChangedDisConnect();
	uiublcw.tW_List->setRowCount(liDomains.count());
	domainListEntry_Sort();
//cout << "qublc: table_Pop - # liDomains entries: " << QString::number(liDomains.count()).toStdString() << endl;                               // *****  debug  *****
	for (i = 0; i < liDomains.count(); ++i) {		// iterate over blocking list and populate widget
		strDomainD = liDomains.at(i);

		for (i2 = 0; i2 < 7; ++i2) {
			switch (i2) {
			    case 0:
				    twItemD = new QTableWidgetItem(strDomainD.dl3);
					twItemD->setTextAlignment(Qt::AlignCenter);
				    break;
			    case 1:
				    twItemD = new QTableWidgetItem(strDomainD.dl2);
					twItemD->setTextAlignment(Qt::AlignCenter);
				    break;
			    case 2:
				    twItemD = new QTableWidgetItem(strDomainD.dltop);
					twItemD->setTextAlignment(Qt::AlignCenter);
				    break;
			    case 3:
				    twItemD = new QTableWidgetItem(strDomainD.include);
					twItemD->setTextAlignment(Qt::AlignCenter);
				    break;
			    case 4:
				    twItemD = new QTableWidgetItem(strDomainD.note);
				    break;
			    case 5:
				    twItemD = new QTableWidgetItem(strDomainD.lastchkd);
					twItemD->setTextAlignment(Qt::AlignCenter);
				    break;
			    case 6:
				    twItemD = new QTableWidgetItem(strDomainD.recfrom);
				    break;
			}
			uiublcw.tW_List->setItem(i, i2, twItemD);
		}
	}
//	uiublcw.tW_List->setCurrentCell(iCurRowIx, iCurCol);
	uiublcw.tW_List->setFocus();
	table_CellChangedConnect();
}

void Qublc::table_HeaderSet()				// set table header
{
	slTH << tr("3rd") << tr("2nd") << tr("top") << tr("include") << tr("note") << tr("last checked") << tr("received while visiting site");
	uiublcw.tW_List->setColumnCount(7);
	uiublcw.tW_List->setHorizontalHeaderLabels(slTH);
	uiublcw.tW_List->setColumnWidth(0, iColWidth0);
	uiublcw.tW_List->setColumnWidth(1, iColWidth1);
	uiublcw.tW_List->setColumnWidth(2, iColWidth2);
	uiublcw.tW_List->setColumnWidth(3, iColWidth3);
	uiublcw.tW_List->setColumnWidth(4, iColWidth4);
	uiublcw.tW_List->setColumnWidth(5, iColWidth5);
	uiublcw.tW_List->setColumnWidth(6, iColWidth6);

	uiublcw.tW_List->setAlternatingRowColors(true);
	uiublcw.tW_List->horizontalHeader()->setStretchLastSection(true);
}

void Qublc::table_HeaderClicked(int iD)		// a table header was clicked - anything to do?
{
	if (iD == 1 or iD == 2) {
		if (iD != iSortDl) {
			iSortDl = iD;
//			domainListEntry_Sort();
			table_Upd();					// sort is done while re-populating widget
		}
	} else {
		mbTxt_Table_HeaderNothingImplemented();
	}
}

void Qublc::table_Upd()						// update widget (pop/clear)
{
	if (liDomains.count() > 0) {
		table_Clear();
		table_Pop();
	} else {
		table_Clear();
	}
}

// *****************************************
// **  about/help/error widget handling   **
// *****************************************
void Qublc::about()							// about clicked
{
	ahw_Create(1);
}

void Qublc::help()							// help clicked
{
	ahw_Create(2);
}

void Qublc::ahw_Create(int iD)				// create about/help/error dialog widget
{
//  cout << "ahwCreate reached " << QString::number( iD ).toStdString() << endl;                                               // *****  debug  *****
	if (bAhw == false) {
		ahw = new aboutHelp();
		connect(this, SIGNAL(sigAhwShow(int)), ahw, SLOT(dlgShow(int)));
		connect(ahw, SIGNAL(sigDelAhw()), this, SLOT(ahw_Del()));
		bAhw = true;
		emit sigAhwShow(iD);
	}
}

void Qublc::ahw_Del()						// ahw widget was closed
{
	ahw->deleteLater();
	bAhw = false;
}

// ********************************************
// **  add new domain entry dialog handling  **
// ********************************************
void Qublc::addDomainDlg_Create()			// create add new domain entry dialog
{
	uiadddomain.setupUi(&wAddDomain);
//	wAddDomain.setAttribute(Qt::WA_DeleteOnClose, true);
	connect(uiadddomain.pB_Add, SIGNAL(clicked()), this, SLOT(addDomainDlg_AddEntry()));
	connect(uiadddomain.pB_Cancel, SIGNAL(clicked()), this, SLOT(addDomainDlg_Hide()));
}

void Qublc::addDomainDlg_Clear()			// clear add new domain entry dialog
{
	uiadddomain.lE_3rdLevel->clear();
	uiadddomain.lE_2ndLevel->clear();
	uiadddomain.lE_TopLevel->clear();
	uiadddomain.cB_Include->setCurrentIndex(1);
	uiadddomain.lE_Note->clear();
	uiadddomain.dE_LastChecked->setDate(QDate::currentDate());
	uiadddomain.lE_ReceivedFrom->clear();
}

void Qublc::addDomainDlg_Hide()				// delete add new domain entry dialog
{
	wAddDomain.hide();
//	wAddDomain.close();
}

void Qublc::addDomainDlg_Show()				// show add new domain entry dialog
{
	addDomainDlg_Clear();
	uiadddomain.lE_TopLevel->setFocus();
	wAddDomain.show();
}

void Qublc::addDomainDlg_AddEntry()			// add new entry from add new domain dialog data
{
	strDomainD.index = 0;
	strDomainD.dl3 = uiadddomain.lE_3rdLevel->text();
	strDomainD.dl2 = uiadddomain.lE_2ndLevel->text();
	strDomainD.dltop = uiadddomain.lE_TopLevel->text();
	strDomainD.include = uiadddomain.cB_Include->currentText();
	strDomainD.note = uiadddomain.lE_Note->text();
	strDomainD.lastchkd = uiadddomain.dE_LastChecked->text();
	strDomainD.recfrom = uiadddomain.lE_ReceivedFrom->text();

	slDomain.clear();
	slDomain.append(strDomainD.dl3);
	slDomain.append(strDomainD.dl2);
	slDomain.append(strDomainD.dltop);
	if (domainListEntry_FindDomain(slDomain) == false) {	// if true, domain already exists
		liDomains.append(strDomainD);
		bListChanged = true;
		addDomainDlg_Hide();
		table_Upd();
	} else {									// domain already exists
		mbTxt_DomainListEntry_DomainExists();
	}
}

// ******************************************************
// **  domain list file functions: create, open, save  **
// ******************************************************
void Qublc::domainListFile_New()			// create new, empty domain list
{
	if (bListChanged == true) {
		iD = mbShow(5, tr("Question - save changes ?"),
		                    tr("The current domain list has been changed.\n \nSave changes before creating a new, empty list?\n"),
		                    tr("Yes - save current list"), tr("No - discard changes"), tr("Cancel - do not create a new domain list"));	// 0=yes, 1=cancel, 2=no
		if (iD == 1) {
			return;
		} else if (iD == 0) {
			domainListFile_Save();
		}
	}
	liDomains.clear();
	uiublcw.l_DomainFile->setText("");
	sFileDomain = "";
	sFileDomainLast = "";
	bListChanged = false;
	table_Upd();
}

void Qublc::domainListFile_OpenDlg()		// file dialog to get domain list
{
	QFileInfo fi(sFileDomainLast);
	QDir d = fi.dir();
	if (d.exists() != true) {
		sFileDomain = QDir::homePath();
	}
	sD = QFileDialog::getOpenFileName(this, tr("Open domain list"), sFileDomainLast, "*.txt");
	if (sD > "") {
		sFileDomain = sD;
		domainListFile_Open();
	} else {
		return;
	}
}

bool Qublc::domainListFile_Open()			// open domain list file
{
	liDomains = rwDomainFileOpen(sFileDomain);		// open domain list file
//cout << "qublc: domainListOpen - # liDomains entries: " << QString::number(liDomains.count()).toStdString() << endl;                                               // *****  debug  *****
	if (liDomains.size() > 1) {						// more than 1 entry
		iD = 0;
		for (i = 0; i < liDomains.count(); ++i) {		// iterate over list and delete invalid entries, count them
			strDomainD = liDomains.at(i);
			if (strDomainD.note == "qublc: err") {			// error, invalid entry - remove
				iD += iD;
				liDomains.removeAt(i);
				i -= i;
			}
		}
		if (liDomains.size() > 0) {
			if (iD > 0) {
				mbTxt_DomainListFileOpen_InvalidEntries(iD);
			}
			uiublcw.l_DomainFile->setText(sFileDomain);
			table_Pop();
			return true;
		} else {
			mbTxt_DomainListFileOpen_NoValidEntry();
			domainListFile_New();
			return false;
		}
	} else if (liDomains.size() == 1) {				// either error opening file or only 1 entry
		strDomainD = liDomains.at(0);
		if (strDomainD.note == "qublc: err") {				// error - either on opening file or the single entry is invalid
			if (strDomainD.dltop == "Error opening file") {		// error - opening file
				mbTxt_DomainListFileOpen_FileOpenError();
				domainListFile_New();
				return false;
			} else {											// error - only entry is invalid
				mbTxt_DomainListFileOpen_NoValidEntry();
				domainListFile_New();
				return false;
			}
		} else {
			uiublcw.l_DomainFile->setText(sFileDomain);
			table_Pop();										// no error, there is one valid entry in the file, populate table widget
			return true;
		}
	} else {										// empty file - liDomains.size() must be 0
		mbTxt_DomainListFileOpen_Empty();
		domainListFile_New();
		return false;
	}
	return false;
}

void Qublc::domainListFile_Save()			// save domain list
{
	if (liDomains.count() < 1) {
		mbTxt_DomainListFileSave_Empty();		// show message box nothing to do, list is empty
		return;
	}

	if (bListChanged == true) {
		if (sFileDomain == "") {
			domainListFile_SaveAs();				// no file name given - call Save As
			return;
		}
		if (domainListFile_Write(0, sFileDomain) == true) {
			bListChanged = false;
			sFileDomainLast = sFileDomain;
		} else {
			mbShow(2, tr("Warning - save failed !"),
			                tr("Saving the domain list seems to be unsuccessful!\n"
			                   "Often this is an access right issue.\n"
			                   "If you have an idea, correct the issue and re-try...\n"
			                   "Else you may try to use 'Save As' and save to a different location...\n"
			                   "Otherwise all changes will be lost on exit!"), tr("&OK"), "", "");
		}
	} else {
		mbTxt_DomainListFileSave_NotChanged();		// show message box nothing to do, list not modified
	}
}

void Qublc::domainListFile_SaveAs()			// save domain list as
{
	if (liDomains.count() < 1) {
		mbTxt_DomainListFileSave_Empty();		// show message box nothing to do, list is empty
		return;
	}

	if (sFileDomainLast != "") {
		sFileDomainAs = sFileDomainLast;		// having a domain list file path+name as default to base of
	} else {
		sFileDomainAs = QDir::homePath();		// use user's home path as default
	}
	QFileInfo fi(sFileDomainAs);
	QDir d = fi.dir();
	if (d.exists() != true) {
		sFileDomainAs = QDir::homePath();
	}
	sFileDomainAs = QFileDialog::getSaveFileName(this, tr("Save domain list as..."), fi.absolutePath(), "*.txt");	// Save As... file dialog
	if (sFileDomainAs > "") {
		if (domainListFile_Write(1, sFileDomainAs) == true) {
			uiublcw.l_DomainFile->setText(sFileDomainAs);
			sFileDomain = sFileDomainAs;
			bListChanged = false;
			sFileDomainLast = sFileDomain;
		}
	}
}

bool Qublc::domainListFile_Write(int iD, const QString& sFile)	// save domain list to file and handle return info, iD: 0=save 1=save as
{
	slD = rwDomainFileSave(sFile, liDomains);
	if (slD.at(0) == "0") {
		mbShow(1, tr("domain list saved!"), slD.at(1), tr("&OK"), "", "");
		return true;
	} else if (slD.at(0) == "99") {
		mbShow(2, tr("Error while saving the domain list!"), slD.at(1), tr("&OK"), "", "");
		return false;
	} else {
		mbShow(2, tr("Error while saving the domain list!"),
		          tr("An unhandled error has occurred.\nThis should not happen.\n"
		             "The domain list is probably NOT saved!"), tr("&OK"), "", "");
		return false;
	}
	return false;
}

// *******************************************************************
// **  domain list entry functions: delete, checks, sort list, upd  **
// *******************************************************************
bool Qublc::domainListEntry_FindDomain(const QStringList& slD)	// check on input that a new domain (0=3rd, 1=2nd, 2=top level domain field) does not already exists - true == exists
{
	if (slD.length() != 3) {
		mbTxt_DomainListEntry_FindDomainErr();
		return true;
	}
	sDomain3rd = slD.at(0);						// 3rd level domain
	sDomain2nd = slD.at(1);						// 2nd level domain
	sDomainTop = slD.at(2);						// top level domain
	for (int i = 0; i < liDomains.count(); ++i) {
		strDomainD2 = liDomains.at(i);
		if (strDomainD2.dltop == sDomainTop and strDomainD2.dl2 == sDomain2nd and strDomainD2.dl3 == sDomain3rd) {
			iDomainListPos = i;
			strDomainMatch = strDomainD2;
			return true;
		}
	}
	return false;
}

bool Qublc::domainListEntry_ChkSeperator(QString& sD)	// check on input that the input text does not contain '|' , the data field seperator in the domain file
{
	if (sD.contains("|")) {
		return true;
	} else {
		return false;
	}
}

void Qublc::domainListEntry_Del()			// delete current domain entry
{
	iCurCellCol = uiublcw.tW_List->currentColumn();
	iCurCellRow = uiublcw.tW_List->currentRow();

//cout << "domainListEntryDel: " << QString::number(iCurCellCol).toStdString() << ";" << QString::number(iCurCellRow).toStdString() << endl;                                  // *****  debug  *****

	sD = uiublcw.tW_List->item(iCurCellRow, 2)->text();		// top level domain
	sD2 = uiublcw.tW_List->item(iCurCellRow, 1)->text();	// 2nd level domain
	sD3 = uiublcw.tW_List->item(iCurCellRow, 0)->text();	// 3rd level domain

//cout << "domainListEntryDel: " << sD.toStdString() << ";" << sD2.toStdString() << ";" << sD3.toStdString() << endl;                                  // *****  debug  *****

	liTwItemsD = uiublcw.tW_List->selectedItems();
	if (liTwItemsD.count() == 1) {	// more than 1 cell selected - abort
		if (mbTxt_DomainListEntry_DelConfirm() == QMessageBox::AcceptRole) {
			if (domainListEntry_DelDo() == false) {
				mbTxt_DomainListEntry_DelErr();
				return;										// error deleting entry in internal list - stop here
			} else {
				uiublcw.tW_List->removeRow(iCurCellRow);	// remove entry from table
				bListChanged = true;
			}
		}
	} else {
		mbTxt_DomainListEntry_DelErrRange();					// error, range detected, more than 1 active cell - currently not supported
	}
}

bool Qublc::domainListEntry_DelDo()			// delete a domain list entry - we assume there are no double entries
{
	for (int i = 0; i < liDomains.count(); ++i) {
		strDomainD = liDomains.at(i);
		if (strDomainD.dltop == sD and strDomainD.dl2 == sD2 and strDomainD.dl3 == sD3) {
			liDomains.removeAt(i);
			return true;
		}
	}
	return false;
}

void Qublc::domainListEntry_Sort()			// sort domain list liDomains - 1st top level, then 2nd level then 3rd - worst sorting algorithm ever - sigh!
{
//cout << "qublc: domainListEntry_Sort - iSortDl = " << QString::number(iSortDl).toStdString() << endl;                               // *****  debug  *****
	iCnt = liDomains.count();
	if (iSortDl == 1) {
		domainListEntry_SortDlL1();				// 1 = 2nd+3rd level domains
		domainListEntry_SortDl3();
	} else if (iSortDl == 2) {
		domainListEntry_SortDlL1();				// 2 = top+2nd+3rd level domains
		domainListEntry_SortDl2();
		domainListEntry_SortDl3();
	} else {
	}
}

void Qublc::domainListEntry_SortDlL1()		// sort domain list liDomains - either according top level (2) domain or 2nd level (1) domain entries
{

	for (i = 1; i < iCnt; ++i) {				// top level domain sort - check every list entry, start with 2nd entry
		strDomainD = liDomains.at(i);
		bBreak = false;
		for (i2 = 0; i2 < i; ++i2) {					// loop over already sorted entries, start with 1st entry, index 0
			strDomainD2 = liDomains.at(i2);
			switch (iSortDl) {
			    case 1:
//cout << "qublc: domainListEntry_SortDlL1 - case 1" << endl;                               // *****  debug  *****
                    if (strDomainD2.dl2 > strDomainD.dl2) {
						liDomains.move(i, i2);
						bBreak = true;
					}
				    break;
			    case 2:
				    if (strDomainD2.dltop > strDomainD.dltop) {
						liDomains.move(i, i2);
						bBreak = true;
					}
				    break;
			    default:
				    break;
			}
			if (bBreak == true) {
				break;
			}
		}
	}
}

void Qublc::domainListEntry_SortDl2()		// sort domain list liDomains - 2nd level domain entries - get top level domain range(s)
{
	iBegin = 0;											// check every list entry, start with 2nd entry
	iEnd = 0;
	strDomainD = liDomains.at(0);
	sPrev = strDomainD.dltop;
	for (i = 1; i < iCnt; ++i) {
		strDomainD = liDomains.at(i);
		sCur = strDomainD.dltop;
		if (sCur == sPrev) {
		} else {
			iEnd = i - 1;
			if (iEnd > iBegin) {						// we have a top level domain range, sort 2nd and 3rd level of this range
				domainListEntry_SortDl2Do();
			} else {									// single top level domain entry, no need to sort
			}
			iBegin = i;
			iEnd = i;
			sPrev = sCur;
		}
	}
	if (iCnt > 1 and iEnd == 0) {						// corner case, all list entries have same top level domain
		iEnd = iCnt - 1;
		domainListEntry_SortDl2Do();
	}
}

void Qublc::domainListEntry_SortDl2Do()		// sort domain list liDomains - 2nd level domain entries per top level domain range(s)
{
	for (i2 = iBegin + 1; i2 < iEnd + 1; ++i2) {		// 2nd level domain sort - check every list entry, start with 2nd entry
		strDomainD = liDomains.at(i2);
		for (i3 = iBegin; i3 < iEnd + 1; ++i3) {		// loop over already sorted entries, start with 1st entry,
			strDomainD2 = liDomains.at(i3);
			if (strDomainD2.dl2 > strDomainD.dl2) {
				liDomains.move(i2, i3);
				break;
			}
		}
	}
}

void Qublc::domainListEntry_SortDl3()		// sort domain list liDomains - 3rd level domain entries - get top + 2nd level domain range(s)
{
	iBegin = 0;											// check every list entry, start with 2nd entry
	iEnd = 0;
	strDomainD = liDomains.at(0);
	sPrev = strDomainD.dltop + strDomainD.dl2;
	for (i = 1; i < iCnt; ++i) {
		strDomainD = liDomains.at(i);
		sCur = strDomainD.dltop + strDomainD.dl2;
		if (sCur == sPrev) {
		} else {
			iEnd = i - 1;
			if (iEnd > iBegin) {						// we have a top and 2nd level domain range, sort 3rd level of this range
				domainListEntry_SortDl3Do();
			} else {									// single top level domain entry, no need to sort
			}
			iBegin = i;
			iEnd = i;
			sPrev = sCur;
		}
	}
	if (iCnt > 1 and iEnd == 0) {						// corner case, all list entries have same top level domain
		iEnd = iCnt - 1;
		    domainListEntry_SortDl3Do();
	}
}

void Qublc::domainListEntry_SortDl3Do()		// sort domain list liDomains - 3rd level domain entries per top + 2nd level domain range(s)
{
	for (i2 = iBegin + 1; i2 < iEnd + 1; ++i2) {		// 2nd level domain sort - check every list entry, start with 2nd entry
		strDomainD = liDomains.at(i2);
		for (i3 = iBegin; i3 < iEnd + 1; ++i3) {		// loop over already sorted entries, start with 1st entry,
			strDomainD2 = liDomains.at(i3);
			if (strDomainD2.dl3 > strDomainD.dl3) {
				liDomains.move(i2, i3);
				break;
			}
		}
	}
}

// ***************************************************
// **  save unbound file							**
// ***************************************************
void Qublc::unboundFile_Create()			// save domain list to unbound readable file
{
	if (liDomains.count() == 0) {				// no entry
		mbTxt_UnboundFile_NoDomains();
		return;
	}

	bD = false;										// stop if not a single entry is to be included
	for (int i = 0; i < liDomains.count(); ++i) {
		strDomainD = liDomains.at(i);
		if (strDomainD.include == "x") {
			bD = true;
			break;
		}
	}
	if (bD == false) {
		mbTxt_UnboundFile_NoInclude();
		return;
	}

	if (sFileUnbound != "") {
		sFileUnboundAs = sFileUnbound;			// having a domain list file path+name as default to base of
	} else {
		sFileUnboundAs = QDir::homePath();		// use user's home path as default
	}
	QFileInfo fi(sFileUnboundAs);
	QDir d = fi.dir();
	if (d.exists() != true) {
		sFileUnboundAs = QDir::homePath();
	}
	sFileUnboundAs = QFileDialog::getSaveFileName(this, tr("Export list as unbound readable file..."), fi.absolutePath(), "*.conf");	// Save As... file dialog
	if (sFileUnboundAs > "") {
		if (unboundFile_Write(1, sFileUnboundAs) == true) {
			sFileUnbound = sFileUnboundAs;
		}
	}
}

bool Qublc::unboundFile_Write(int iD, const QString& sFile)	// write domain list to unbound readable file and handle return info, iD: 0=save 1=save as
{
	slD = rwUnboundFileSave(sFile, liDomains);
	if (slD.at(0) == "0") {
		mbShow(1, tr("unbound blacklist *.conf file saved!"), slD.at(1), tr("&OK"), "", "");
		return true;
	} else if (slD.at(0) == "99") {
		mbShow(2, tr("Error while saving the unbound blacklist *.conf file!"), slD.at(1), tr("&OK"), "", "");
		return false;
	} else {
		mbShow(2, tr("Error while saving the unbound blacklist *.conf!"),
		                tr("An unhandled error has occurred.\nThis should not happen.\n"
		                    "The unbound blacklist *.conf file could most likely not be created!"), tr("&OK"), "", "");
		return false;
	}
	return false;
}

// ****************************************************
// **  message box and it's messages                 **
// ****************************************************
int Qublc::mbShow(int iD, const QString& sD, const QString& sD2, const QString& sD3, const QString& sD4, const QString& sD5)
{						// show QMessageBox at cursor position: type, title, text, button accept, button reject, button destruct
	QMessageBox msgBox;
	msgBox.setWindowTitle(sD);

	if (iD == 1) {
		msgBox.setIcon(QMessageBox::Information);
	} else if (iD == 2) {
		msgBox.setIcon(QMessageBox::Warning);
	} else if (iD == 3) {
		msgBox.setIcon(QMessageBox::Critical);
	} else if (iD == 4) {
		msgBox.setIcon(QMessageBox::Question);	// 2 buttons
	} else if (iD == 5) {
		msgBox.setIcon(QMessageBox::Question);	// 3 buttons, 0=yes, 1=cancel, 2=no
	} else {
		return 9;
	}

	msgBox.setText(sD2);

	msgBox.addButton(sD3, QMessageBox::AcceptRole);
	if (iD == 5) {
		msgBox.addButton(sD5, QMessageBox::DestructiveRole);
	}
	if (iD > 3) {
		msgBox.addButton(sD4, QMessageBox::RejectRole);
	}

	iX = iScrX/2 - msgBox.frameGeometry().width()/2;		// this adjustment should center the mbShow - but it doesn't !?
	if (iX < 0) {
		iX = 0;
	}
	iY = iScrY/2 - msgBox.frameGeometry().height()/2;
	if (iY < 0) {
		iY = 0;
	}
	msgBox.move(iX, iY);

	iRet = msgBox.exec();
	return iRet;
}

void Qublc::mbTxt_DomainListEntry_ChgReverted()			// show message box - this domain already exists, revert change - shown when table cells were changed
{
	mbShow(1, tr("Information - Domain already in list !"),
	          tr("This domain already exists in the list.\n"
	             "Your change will be reverted to avoid double domain entries in the list."), tr("&OK"), "", "");
}

int Qublc::mbTxt_DomainListEntry_DelConfirm()			// show message box - confirm deletion of list entry
{
	return mbShow(4, tr("Question - sure to delete entry ?"),
	          tr("Are you sure to delete this domain list entry?"), tr("&Yes - delete"), "&No - keep", "");
}

void Qublc::mbTxt_DomainListEntry_DelErr()				// show message box - error while trying to delete a domain list entry
{
	mbShow(2, tr("Warning - error while deleting entry !"),
	          tr("An unexpected error occurred while trying to delete the selected entry in the internal list.\n"
	             "This list feeds the table and it's content is saved as a file!\n\n"
	             "This must never happen!?!\n"
	             "It seems, that the list and the table are out of sync!\n"
	             "The entry will not be deleted from the table.\n\n"
	             "You better close the program and re-load the list from file.\n\n"
	             "Sorry for the inconvenience!"), tr("&OK"), "", "");
}

int Qublc::mbTxt_DomainListEntry_DelErrRange()			// show message box - error, range detected, more than 1 active cell - currently not supported
{
	return mbShow(1, tr("Information - error - multiple active cells !"),
	          tr("There is more than 1 cell active/selected.\n"
	             "Due to the laziness of the author, this is currently not supported,\n"
	             "even if all cells are in one row, belonging to one entry.\n\n"
	             "If you really need to delete a hugh number of entries,\n"
	             "please use a text editor on the file.\n"), tr("&OK"), "", "");
}

void Qublc::mbTxt_DomainListEntry_DomainExists()		// show message box - domain already exists - shown when adding a new domain
{
	mbShow(1, tr("Information - Domain already in list !"),
	          tr("This domain already exists in the list.\n"
	             "Please change or 'Cancel' to avoid double domain entries in the list."), tr("&OK"), "", "");
}

void Qublc::mbTxt_DomainListEntry_InvalidInclude()		// show message box - wrong include input
{
	mbShow(1, tr("Information - Invalid input !"),
	          tr("The include field can only have one of two valid entries:\n"
	             "empty = do not include\n"
	             "x = include\n"
	             "entry when writing the unbound *.conf file.\n"
	             "Field will be re-set, please try again."), tr("&OK"), "", "");
}

void Qublc::mbTxt_DomainListEntry_FindDomainErr()		// show message box - find domain error - domain QStringList != 3 entries
{
	mbShow(2, tr("Warning - Internal error !"),
	          tr("domainListEntry_FindDomain called with QStringList,\n"
	             "which has not exactly 3 entries.\n# entries: %1").arg(QString::number(slD.length())), tr("&OK"), "", "");
}

void Qublc::mbTxt_DomainListEntry_SepRep()				// show message box - the data field separator '|' was entered and replaced by '/'
{
	mbShow(1, tr("Attention - illegal character used !"),
	          tr("The character '|' has been replaced with '/'.\n"
	             "Please check whether this is OK for you or replace it with something else, but not with '|'.\n\n"
	             "'|' is used internally as the data set field separator for a domain entry data set and cannot be used as text."), tr("&OK"), "", "");
}

void Qublc::mbTxt_DomainListFileOpen_Empty()			// show message box - domain list is empty on open
{
	mbShow(1, tr("Information - empty file !"),
	          tr("There is nothing to do.\n"
	             "The domain list file is empty."), tr("&OK"), "", "");
}

void Qublc::mbTxt_DomainListFileOpen_FileOpenError()	// show message box - error - opening file
{
	mbShow(1, tr("Information - error opening file !"),
	          tr("Cannot open file\n"
	             "%1\n"
	             "for reading.").arg(sFileDomain), tr("&OK"), "", "");
}

void Qublc::mbTxt_DomainListFileOpen_InvalidEntries(int iD2)	// show message box - error - ignored invalid entries
{
	mbShow(1, tr("Information - invalid entries in file !"),
	          tr("While opening the file\n"
	             "%1 ,\n"
	             "%2 invalid entries were ignored.\n"
	             "This should never happen!?").arg(sFileDomain).arg(QString::number(iD2)), tr("&OK"), "", "");
}

void Qublc::mbTxt_DomainListFileOpen_NoValidEntry()		// show message box - error - no valid entries
{
	mbShow(1, tr("Information - no valid entries !"),
	          tr("Could not found a single valid entry."), tr("&OK"), "", "");
}

void Qublc::mbTxt_DomainListFileSave_Empty()			// show message box - domain list is empty on save or save as
{
	mbShow(1, tr("Information - nothing to save !"),
	          tr("There is nothing to do.\n"
	             "The domain list is currently empty."), tr("&OK"), "", "");
}

void Qublc::mbTxt_DomainListFileSave_NotChanged()		// show message box - domain list was not changed on save or save as
{
	mbShow(1, tr("Information - nothing to save !"),
	          tr("There is nothing to do.\n"
	             "The list has not been modified."), tr("&OK"), "", "");
}

void Qublc::mbTxt_Table_HeaderNothingImplemented()		// show message box - table header click not assigned to anything
{
	mbShow(1, tr("Information - nothing implemented !"),
	          tr("Currently only clicking on the table headers:\n"
	             "- '2nd', will sort according 2nd and 3rd domain levels, and\n"
	             "- '3rd', will sort according top, 2nd and 3rd domain levels\n"
	             "will do something."), tr("&OK"), "", "");
}

void Qublc::mbTxt_UnboundFile_NoDomains()				// show message box - no domains - list is empty, no point to create file
{
	mbShow(1, tr("Information - no domains !"),
	          tr("There is nothing to do.\n"
	             "The domain list is empty.\n"
	             "No point to create an empty unbound blacklist.conf file."), tr("&OK"), "", "");
}

void Qublc::mbTxt_UnboundFile_NoInclude()				// show message box - no entry to include in unbound file
{
	mbShow(1, tr("Information - nothing to save !"),
	          tr("No entry is marked to be included in the file for unbound."), tr("&OK"), "", "");
}

//#include "qublc.moc"
